"""Default logger for py_ballisticcalc library"""
import logging

__all__ = ('logger',
           'enable_file_logging',
           'disable_file_logging',
           'get_debug',
           'set_debug',
           )

formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)
console_handler.setLevel(logging.INFO)  # Default level for console

logger = logging.getLogger('py_balcalc')
logger.addHandler(console_handler)

# Global DEBUG variable
DEBUG = False

# File handler (optional, added dynamically)
file_handler = None


def enable_file_logging(filename: str = "debug.log") -> None:
    """Enable logging to a file, replacing any existing file handler."""
    global file_handler
    # Remove the existing file handler if it exists
    if file_handler is not None:
        disable_file_logging()

    # Add a new file handler
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(logging.DEBUG)  # Log everything to the file
    file_formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)


def disable_file_logging() -> None:
    """Disable logging to a file if it is currently enabled."""
    global file_handler
    if file_handler is not None:
        logger.removeHandler(file_handler)
        file_handler.close()
        file_handler = None


def set_debug(value: bool) -> None:
    """Set the global DEBUG variable and adjust logging levels."""
    global DEBUG
    DEBUG = value
    if DEBUG:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


def get_debug() -> bool:
    """Get the current value of the global DEBUG variable."""
    return DEBUG



"""Computes trajectory using SciPy's solve_ivp; uses SciPy's root_scalar to get specific points.
TODO:
 * Use SciPy root finder for zero_angle() override
"""
import math
import warnings
from typing_extensions import Union, Tuple, List, Optional, override

from py_ballisticcalc.conditions import Shot, Wind
from py_ballisticcalc.engines.base_engine import BaseIntegrationEngine, BaseEngineConfigDict, create_trajectory_row
from py_ballisticcalc.exceptions import RangeError
from py_ballisticcalc.logger import logger
from py_ballisticcalc.trajectory_data import TrajectoryData, TrajFlag
from py_ballisticcalc.unit import Distance, PreferredUnits
from py_ballisticcalc.vector import Vector

__all__ = ('SciPyIntegrationEngine',)

def custom_warning_format(message, category, filename, lineno, file=None, line=None):
    return f"{category.__name__}: {message}\n"
warnings.formatwarning = custom_warning_format

class WindSock():
    """Finds wind vector in effect at any distance down-range."""
    def __init__(self, winds: Union[Tuple["Wind", ...], None]):
        # Sort winds by range, ascending
        self.winds = None
        if isinstance(winds, Wind):
            self.winds = [winds]
        elif isinstance(winds, (tuple, list)):
            self.winds = sorted(winds, key=lambda w: w.until_distance.raw_value)

    def wind_at_distance(self, distance: float) -> Optional[Vector]:
        """Returns wind vector at specified distance, where distance is in feet."""
        if not self.winds:
            return None
        for wind in self.winds:  # TODO: use binary search for performance
            distance *= 12.0  # Convert distance to inches
            if distance <= wind.until_distance.raw_value:
                return wind.vector
        return self.winds[-1].vector


#pylint: disable=import-outside-toplevel,unused-argument,too-many-statements
class SciPyIntegrationEngine(BaseIntegrationEngine):
    """Integration engine using SciPy's solve_ivp for trajectory calculations."""

    DEFAULT_MAX_TIME = 90.0  # Max flight time to simulate before stopping integration
    DEFAULT_ERROR_TOLERANCE = 1e-8  # Default relative tolerance (rtol) for integration
    DEFAULT_INTEGRATION_METHOD = 'DOP853'  # Default integration method for solve_ivp
    # Other recommended methods: 'BDF', 'LSODA', 'Radau'

    def __init__(self, config: BaseEngineConfigDict,
        max_time: Optional[float] = None,
        integration_method: Optional[str] = None,
        error_tolerance: Optional[float] = None
    ):
        """
        Initializes the SciPyIntegrationEngine.

        Args:
            config (BaseEngineConfigDict): Configuration dictionary for the engine.
            max_time (float, optional): Maximum time to simulate in seconds. Defaults to DEFAULT_MAX_TIME.
            integration_method (str, optional): Integration method to use with solve_ivp. Defaults to DEFAULT_INTEGRATION_METHOD.
            error_tolerance (float, optional): Relative tolerance for integration. Defaults to DEFAULT_ERROR_TOLERANCE.
        """
        super().__init__(config)
        self.max_time = max_time if max_time is not None else self.DEFAULT_MAX_TIME
        self.integration_method = integration_method if integration_method is not None else self.DEFAULT_INTEGRATION_METHOD
        self.error_tolerance = error_tolerance if error_tolerance is not None else self.DEFAULT_ERROR_TOLERANCE

    @override
    def _integrate(self, shot_info: Shot, maximum_range: float, record_step: float,
                   filter_flags: Union[TrajFlag, int], time_step: float = 0.0) -> List[TrajectoryData]:
        """
        Calculate trajectory for specified shot

        Args:
            shot_info (Shot):  Information about the shot
            maximum_range (float): Feet down range to stop calculation
            record_step (float): Frequency (in feet down range) to record TrajectoryData
            filter_flags (Union[TrajFlag, int]): Bitfield for requesting special trajectory points
            time_step (float, optional): Maximum time (in seconds) between TrajectoryData records

        Returns:
            List[TrajectoryData]: list of TrajectoryData, one for each dist_step, out to max_range
        """

        try:
            from scipy.integrate import solve_ivp
            from scipy.optimize import root_scalar
            import numpy as np
        except ImportError as e:
            raise ImportError("SciPy is required for SciPyIntegrationEngine.") from e

        _cMinimumVelocity = self._config.cMinimumVelocity
        _cMaximumDrop = self._config.cMaximumDrop
        _cMinimumAltitude = self._config.cMinimumAltitude

        ranges: List[TrajectoryData] = []  # Record of TrajectoryData points to return

        wind_sock = WindSock(shot_info.winds)

        #region Initialize velocity and position of projectile
        velocity = self.muzzle_velocity
        # x: downrange distance, y: drop, z: windage
        # s = [x, y, z, vx, vy, vz]
        s0 = [.0, -self.cant_cosine * self.sight_height, -self.cant_sine * self.sight_height,
              math.cos(self.barrel_elevation) * math.cos(self.barrel_azimuth) * velocity,
              math.sin(self.barrel_elevation) * velocity,
              math.cos(self.barrel_elevation) * math.sin(self.barrel_azimuth) * velocity]
        #endregion

        #region SciPy integration
        def diff_eq(t, s):
            """
            Defines the dynamics of the bullet for integration.
            :param t: Time (not used in this case, but required by solve_ivp)
            :param y: State vector [x, y, z, vx, vy, vz]
            :return: Derivative of state vector
            """
            x, y, z = s[:3]         #pylint: disable=unused-variable
            vx, vy, vz = s[3:]
            velocity_vector = Vector(vx, vy, vz)
            wind_vector = wind_sock.wind_at_distance(x)
            if wind_vector is None:
                relative_velocity = velocity_vector
            else:
                relative_velocity = velocity_vector - wind_vector
            relative_speed = relative_velocity.magnitude()
            density_factor, mach = shot_info.atmo.get_density_factor_and_mach_for_altitude(self.alt0 + y)
            km = density_factor * self.drag_by_mach(relative_speed / mach)
            drag = km * relative_speed

            # Derivatives
            dxdt = vx
            dydt = vy
            dzdt = vz
            dvxdt = -drag * relative_velocity.x
            dvydt = self.gravity_vector.y - drag * relative_velocity.y
            dvzdt = -drag * relative_velocity.z
            return [dxdt, dydt, dzdt, dvxdt, dvydt, dvzdt]
        #endregion SciPy integration

        def event_max_range(t, s):  # Stop when x crosses maximum_range
            return s[0] - (maximum_range + 1)  # +1 to ensure we cross the threshold
        event_max_range.terminal = True  # type: ignore[attr-defined]

        max_drop = max(_cMaximumDrop, _cMinimumAltitude - self.alt0)
        def event_max_drop(t, s):  # Stop when y crosses max_drop
            return s[1] - max_drop
        event_max_drop.terminal = True  # type: ignore[attr-defined]
        event_max_drop.direction = -1  # type: ignore[attr-defined]

        def event_min_velocity(t, s):  # Stop when velocity < _cMinimumVelocity
            v = np.linalg.norm(s[3:6])
            return v - _cMinimumVelocity
        event_min_velocity.terminal = True  # type: ignore[attr-defined]

        traj_events = [event_max_range, event_max_drop, event_min_velocity]

        if filter_flags & TrajFlag.ZERO:
            def zero_crossing(t, s):  # Look for trajectory crossing sight line
                # Solve for y = x * tan(look_angle)
                return s[1] - s[0] * math.tan(self.look_angle)
            zero_crossing.terminal = False  # type: ignore[attr-defined]
            zero_crossing.direction = 0  # type: ignore[attr-defined]
            traj_events.append(zero_crossing)

        sol = solve_ivp(diff_eq, (0, self.max_time), s0,
                        method=self.integration_method, dense_output=True, rtol=self.error_tolerance,
                        events=traj_events)

        if not sol.success:  # Integration failed
            raise RangeError(f"SciPy integration failed: {sol.message}", ranges)

        logger.debug(f"SciPy integration complete with {sol.nfev} function calls.")
        termination_reason = None
        if sol.status == 1:  # A termination event occurred
            if len(sol.t_events) > 0:
                # if sol.t_events[0].size > 0:  # Typical termination event
                #     logger.debug(f"Integration stopped at max range: {sol.t_events[0][0]}")
                if sol.t_events[1].size > 0:  # event_max_drop
                    y = sol.sol(sol.t_events[1][0])[1]  # Get y at max drop event
                    if y < _cMaximumDrop:
                        termination_reason = RangeError.MaximumDropReached
                    else:
                        termination_reason = RangeError.MinimumAltitudeReached
                elif sol.t_events[2].size > 0:  # event_min_velocity
                    termination_reason = RangeError.MinimumVelocityReached

        #region Find requested TrajectoryData points
        if sol.sol is not None and sol.status != -1:
            def make_row(t: float, state: np.ndarray, flag: Union[TrajFlag, int]) -> TrajectoryData:
                """Helper function to create a TrajectoryData row."""
                position = Vector(*state[0:3])
                velocity = Vector(*state[3:6])
                velocity_magnitude = velocity.magnitude()
                density_factor, mach = shot_info.atmo.get_density_factor_and_mach_for_altitude(self.alt0 + position[1])
                drag = density_factor * velocity_magnitude * self.drag_by_mach(velocity_magnitude / mach)
                return create_trajectory_row(t, position, velocity, velocity_magnitude, mach,
                    self.spin_drift(t), self.look_angle, density_factor, drag, self.weight, flag
                )

            if sol.t[-1] == 0:
                # If the last time is 0, we only have the initial state
                ranges.append(make_row(sol.t[0], sol.y[:, 0], TrajFlag.RANGE))
                return ranges

            # List of distances at which we want to record the trajectory data
            desired_xs = np.arange(0, maximum_range + record_step, record_step)
            # Get x and t arrays from the solution
            x_vals = sol.y[0]
            t_vals = sol.t
            # Interpolate to find the time when each desired x is reached
            t_at_x = np.interp(desired_xs, x_vals, t_vals)

            #region Basic approach to interpolate for desired x values:
            # # This is not very good: distance from requested x varies by ~1% of max_range
            # states_at_x = sol.sol(t_at_x)  # shape: (state_dim, len(desired_xs))
            # for i in range(states_at_x.shape[1]):
            #     ranges.append(make_row(t_at_x[i], states_at_x[:, i], TrajFlag.RANGE))
            #endregion Basic approach to interpolate for desired x values

            #region Root-finding approach to interpolate for desired x values:
            warnings.simplefilter("once")  # Only issue one warning
            states_at_x = []
            t_at_x = []
            for x_target in desired_xs:
                idx = np.searchsorted(x_vals, x_target)  # Find bracketing indices for x_target
                if idx < 0 or idx >= len(x_vals):
                    warnings.warn(f"Requested range exceeds computed trajectory, which only reaches {PreferredUnits.distance(Distance.Feet(x_vals[-1]))}",
                                  RuntimeWarning)
                    continue
                if idx == 0:
                    if filter_flags == TrajFlag.NONE:
                        continue  # Don't record first point
                    t_root = t_vals[0]
                else:
                    t_lo, t_hi = t_vals[idx - 1], t_vals[idx]
                    # Use root_scalar to find t where x(t) == x_target
                    def x_minus_target(t):  # Function for root finding: x(t) - x_target
                        return sol.sol(t)[0] - x_target  # pylint: disable=cell-var-from-loop
                    res = root_scalar(x_minus_target, bracket=[t_lo, t_hi], method='brentq') #, xtol=1e-14, rtol=1e-14)
                    # #region Newton's method to find root
                    # def dxdt(t):
                    #     return sol.sol(t)[3]  # vx(t)
                    # res = root_scalar(x_minus_target, fprime=dxdt, bracket=[t_lo, t_hi],
                    #                   x0=0.5 * (t_lo + t_hi),  # Initial guess for Newton's method
                    #                   method='newton', xtol=1e-14, rtol=1e-14)
                    # #endregion Newton's method to find root
                    if not res.converged:
                        logger.warning(f"Could not find root for requested distance {x_target}")
                        continue
                    t_root = res.root
                state = sol.sol(t_root)
                t_at_x.append(t_root)
                states_at_x.append(state)
            #region Root-finding approach to interpolate for desired x values

            # If we ended with an event then also grab the last point calculated
            if termination_reason is not None and len(t_vals) > 1 and t_vals[0] != t_vals[-1]:
                t_at_x.append(sol.t[-1])  # Last time point
                states_at_x.append(sol.y[:, -1])  # Last state at the end of integration

            t_at_x = np.array(t_at_x)  # Convert to arrays for easier handling
            states_at_x = np.array(states_at_x).T  # shape: (state_dim, num_points)
            for i in range(states_at_x.shape[1]):
                ranges.append(make_row(t_at_x[i], states_at_x[:, i], TrajFlag.RANGE))
            ranges.sort(key=lambda t: t.time)  # Sort by time

            if time_step > 0.0:
                time_of_last_record = 0.0
                for next_record in range(1, len(ranges)):
                    while ranges[next_record].time - time_of_last_record > time_step:
                        time_of_last_record += time_step
                        ranges.append(make_row(time_of_last_record, sol.sol(time_of_last_record), TrajFlag.RANGE))
                ranges.sort(key=lambda t: t.time)  # Sort by time

            #region Find TrajectoryData points requested by filter_flags
            if filter_flags:
                if filter_flags & TrajFlag.MACH and ranges[0].mach > 1.0 and ranges[-1].mach < 1.0:
                    def mach_minus_one(t):
                        """Returns the Mach number at time t minus 1."""
                        state = sol.sol(t)
                        x, y = state[:2]
                        relative_velocity = Vector(*state[3:])
                        if (wind_vector := wind_sock.wind_at_distance(x)) is not None:
                            relative_velocity = relative_velocity - wind_vector
                        relative_speed = relative_velocity.magnitude()
                        _, mach = shot_info.atmo.get_density_factor_and_mach_for_altitude(self.alt0 + y)
                        return (relative_speed / mach) - 1.0
                    try:
                        res = root_scalar(mach_minus_one, bracket=[t_vals[0], t_vals[-1]])
                        if res.converged:
                            ranges.append(make_row(res.root, sol.sol(res.root), TrajFlag.MACH))
                    except ValueError:
                        logger.debug("No Mach crossing found")

                if filter_flags & TrajFlag.ZERO and len(sol.t_events) > 3 and sol.t_events[-1].size > 0:
                    tan_look_angle = math.tan(self.look_angle)
                    for t_cross in sol.t_events[-1]:
                        state = sol.sol(t_cross)
                        # To determine crossing direction, sample after the crossing
                        dt = 1e-8  # Small time offset
                        state_after = sol.sol(t_cross + dt)
                        y_after = state_after[1] - state_after[0] * tan_look_angle
                        if y_after > 0:
                            direction = TrajFlag.ZERO_UP
                        else:
                            direction = TrajFlag.ZERO_DOWN
                        ranges.append(make_row(t_cross, state, direction))

                if filter_flags & TrajFlag.APEX:
                    def vy(t):
                        """Returns the vertical velocity at time t."""
                        return sol.sol(t)[4]
                    try:
                        res = root_scalar(vy, bracket=[t_vals[0], t_vals[-1]])
                        if res.converged:
                            ranges.append(make_row(res.root, sol.sol(res.root), TrajFlag.APEX))
                    except ValueError:
                        logger.debug("No apex found for trajectory")

                ranges.sort(key=lambda t: t.time)  # Sort by time
                #endregion Find TrajectoryData points requested by filter_flags
            #endregion Find requested TrajectoryData points

            if termination_reason is not None:
                raise RangeError(termination_reason, ranges)
        else:
            logger.error("No solution found by SciPy integration.")
        #endregion Process the solution
        return ranges



"""
This module defines the `EngineProtocol`, a protocol that outlines the
interface for ballistic trajectory calculation engines within the
py_ballisticcalc library.

It specifies the methods that concrete engine implementations must provide
to perform trajectory calculations, access drag model data, and determine
zeroing angles for firearms.
"""

from typing import TypeVar

from typing_extensions import List, Protocol, runtime_checkable

from py_ballisticcalc.conditions import Shot
from py_ballisticcalc.drag_model import DragDataPoint
from py_ballisticcalc.trajectory_data import TrajectoryData
from py_ballisticcalc.unit import Distance, Angular

ConfigT = TypeVar("ConfigT", covariant=True)


@runtime_checkable
class EngineProtocol(Protocol[ConfigT]):
    """
    Defines the interface for a ballistic trajectory calculation engine.

    This protocol outlines the methods that any concrete ballistic engine
    implementation should provide to perform trajectory calculations,
    retrieve drag model information, and determine zeroing angles.
    """

    def __init__(self, _config: ConfigT):
        """
        Initializes the TrajectoryCalc class.

        Args:
            _config (Config): The configuration object.
        """

    @property
    def table_data(self) -> List[DragDataPoint]:
        """
        Gets the drag model table data.

        Returns:
            List[DragDataPoint]: A list of drag data points.
        """

    def trajectory(self, shot_info: Shot, max_range: Distance, dist_step: Distance,
                   extra_data: bool = False, time_step: float = 0.0) -> List[TrajectoryData]:
        """
        Calculates the trajectory of a projectile.

        Args:
            shot_info (Shot): Information about the shot.
            max_range (Distance): The maximum range of the trajectory.
            dist_step (Distance): The distance step for calculations.
            extra_data (bool, optional): Flag to include extra data. Defaults to False.
            time_step (float, optional): The time step for calculations. Defaults to 0.0.

        Returns:
            List[TrajectoryData]: A list of trajectory data points.
        """

    def zero_angle(self, shot_info: Shot, distance: Distance) -> Angular:
        """
        Iterative algorithm to find barrel elevation needed for a particular zero

        Args:
            shot_info (Shot): Shot parameters
            distance (Distance): Zero distance

        Returns:
            Angular: Barrel elevation to hit height zero at zero distance
        """



"""Global Constants"""

from typing_extensions import Final

# Global atmosphere constants
cStandardHumidity: Final[float] = 0.0  # Relative Humidity
cPressureExponent: Final[float] = 5.255876  # =g*M/R*L

# ISA, metric prefer_units: (https://www.engineeringtoolbox.com/international-standard-atmosphere-d_985.html)
cDegreesCtoK: Final[float] = 273.15  # °K = °C + 273.15
cStandardTemperatureC: Final[float] = 15.0  # °C
cLapseRateKperFoot: Final[float] = -0.0019812  # Lapse Rate, °K/ft
cLapseRateMetric: Final[float] = -6.5e-03  # Lapse Rate, °C/m
cStandardPressureMetric: Final[float] = 1013.25  # hPa
cSpeedOfSoundMetric: Final[float] = 20.0467  # Mach1 in m/s = cSpeedOfSound * sqrt(°K)
cStandardDensityMetric: Final[float] = 1.2250  # kg/m^3
cDensityImperialToMetric: Final[float] = 16.0185  # lb/ft^3 to kg/m^3

# ICAO standard atmosphere:
cDegreesFtoR: Final[float] = 459.67  # °R = °F + 459.67
cStandardTemperatureF: Final[float] = 59.0  # °F
cLapseRateImperial: Final[float] = -3.56616e-03  # Lapse rate, °F/ft
cStandardPressure: Final[float] = 29.92  # InHg
cSpeedOfSoundImperial: Final[float] = 49.0223  # Mach1 in fps = cSpeedOfSound * sqrt(°R)
cStandardDensity: Final[float] = 0.076474  # lb/ft^3

# Runtime constants
cLowestTempF: Final[float] = -130  # °F
cMaxWindDistanceFeet: Final[float] = 1e8



# type: ignore
# pylint: skip-file

import math
import warnings

from typing_extensions import Optional

from py_ballisticcalc import HitResult
from py_ballisticcalc.helpers import find_time_for_distance_in_shot
from py_ballisticcalc.trajectory_data import TrajFlag, DangerSpace
from py_ballisticcalc.unit import PreferredUnits, Angular

try:
    import matplotlib
    from matplotlib.patches import Polygon
    from matplotlib.axes import Axes
    from matplotlib import pyplot, ticker
except ImportError as error:
    warnings.warn("Install matplotlib to get results as a plot", UserWarning)
    raise error

__all__ = (
    'show_hit_result_plot',
    'add_danger_space_overlay',
    'hit_result_as_plot',
)

PLOT_FONT_HEIGHT = 72
PLOT_FONT_SIZE = 552 / PLOT_FONT_HEIGHT

PLOT_COLORS = {
    "trajectory": (130 / 255, 179 / 255, 102 / 255, 1.0),
    "frame": (.0, .0, .0, 1.0),
    "velocity": (108 / 255, 142 / 255, 191 / 255, 1.0),
    "sight": (150 / 255, 115 / 255, 166 / 255, 1.0),
    "barrel": (184 / 255, 84 / 255, 80 / 255, 1.0),
    "face": (.0, .0, .0, .0),
    TrajFlag.ZERO_UP: (215 / 255, 155 / 255, .0),
    TrajFlag.ZERO_DOWN: (108 / 255, 142 / 255, 191 / 255, 1.0),
    TrajFlag.MACH: (184 / 255, 84 / 255, 80 / 255, 1.0)
}


def show_hit_result_plot() -> None:
    pyplot.show()


def add_danger_space_overlay(danger_space: DangerSpace, ax: 'Axes', label: Optional[str] = None):  # type: ignore
    """Highlights danger-space region on plot"""

    cosine = math.cos(danger_space.look_angle >> Angular.Radian)
    begin_dist = (danger_space.begin.distance >> PreferredUnits.distance) * cosine
    begin_drop = (danger_space.begin.height >> PreferredUnits.drop) * cosine
    end_dist = (danger_space.end.distance >> PreferredUnits.distance) * cosine
    end_drop = (danger_space.end.height >> PreferredUnits.drop) * cosine
    range_dist = (danger_space.at_range.distance >> PreferredUnits.distance) * cosine
    range_drop = (danger_space.at_range.height >> PreferredUnits.drop) * cosine
    h = danger_space.target_height >> PreferredUnits.drop

    # Target position and height:
    ax.plot((range_dist, range_dist), (range_drop + h / 2, range_drop - h / 2),
            color='r', linestyle=':')
    # Shaded danger-space region:
    vertices = (
        (begin_dist, begin_drop), (end_dist, begin_drop),
        (end_dist, end_drop), (begin_dist, end_drop)
    )
    polygon = Polygon(vertices, closed=True,
                      edgecolor='none', facecolor='r', alpha=0.3)
    ax.add_patch(polygon)
    if label is None:  # Add default label
        label = f"Danger space\nat {danger_space.at_range.distance << PreferredUnits.distance}"
    if label != '':
        ax.text(begin_dist + (end_dist - begin_dist) / 2, end_drop, label, color='r',
                linespacing=1.2, fontsize=PLOT_FONT_SIZE, ha='center', va='top')


def add_time_of_flight_axis(ax: 'Axes', hit_result: HitResult, time_precision: int = 1) -> 'Axes':
    """Add on the top of shot plot the axis, corresponding to time of flight in seconds.
       The ticks used for time axis correspond to distance ticks at the bottom of the plot, where it makes sense.
       :param time_precision: describes how much decimal points should be used in formatting time values
    """

    def time_label_for_distance(
            distance_in_unit, distance_unit, decimal_point_in_seconds
    ):
        time = find_time_for_distance_in_shot(hit_result, distance_in_unit, distance_unit)
        if math.isnan(time):
            return ""
        else:
            return f"{time:.{decimal_point_in_seconds}f}"

    twin_x_axes = ax.twiny()
    twin_x_axes.set_xlim(ax.get_xlim())
    xticks = ax.get_xticks()
    shot_last_distance = hit_result[-1].distance >> PreferredUnits.distance
    sensible_time_ticks = xticks[xticks <= shot_last_distance]
    sensible_time_ticks = sensible_time_ticks[sensible_time_ticks >= 0]

    sensible_top_labels = [
        time_label_for_distance(x, PreferredUnits.distance, time_precision)
        for x in sensible_time_ticks
    ]
    twin_x_axes.xaxis.set_major_locator(ticker.FixedLocator(sensible_time_ticks))
    twin_x_axes.xaxis.set_major_formatter(ticker.FixedFormatter(sensible_top_labels))

    twin_x_axes.set_xlabel("s")
    return ax


def hit_result_as_plot(hit_result, look_angle: Optional[Angular] = None,
                       show_time_axis: bool = True) -> 'Axes':  # type: ignore
    """
    :param hit_result: HitResult object
    :param look_angle: look_angle
    :param show_time_axis: whether time of flight axis should be added to plot
    :return: graph of the trajectory
    """

    if look_angle is None:
        look_angle = hit_result.shot.look_angle

    if not hit_result.extra:
        warnings.warn("HitResult.plot: To show extended data"
                      "Use Calculator.fire(..., extra_data=True)")

    font_size = PLOT_FONT_SIZE
    df = hit_result.dataframe()
    fig, ax = pyplot.subplots()

    ax = df.plot(x='distance', y=['height'], ylabel=PreferredUnits.drop.symbol,
                 color=PLOT_COLORS['trajectory'], linewidth=2, ax=ax)
    # max_range = hit_result.trajectory[-1].distance >> PreferredUnits.distance  # This doesn't correctly handle backward-bending trajectories
    max_range = df.distance.max()
    backward_bending_trajectory = (hit_result[-1].distance >> PreferredUnits.distance) != max_range

    for p in hit_result.trajectory:
        if TrajFlag(p.flag) & TrajFlag.ZERO:
            ax.plot([p.distance >> PreferredUnits.distance, p.distance >> PreferredUnits.distance],
                    [df['height'].min(), p.height >> PreferredUnits.drop], linestyle=':',
                    color=PLOT_COLORS[TrajFlag(p.flag) & TrajFlag.ZERO])
            ax.text((p.distance >> PreferredUnits.distance) + max_range / 100, df['height'].min(),
                    f"{TrajFlag.name(TrajFlag(p.flag) & TrajFlag.ZERO)}",
                    fontsize=font_size, rotation=90, color=PLOT_COLORS[TrajFlag(p.flag) & TrajFlag.ZERO])
        if TrajFlag(p.flag) & TrajFlag.MACH:
            ax.plot([p.distance >> PreferredUnits.distance, p.distance >> PreferredUnits.distance],
                    [df['height'].min(), p.height >> PreferredUnits.drop],
                    linestyle=':', color=PLOT_COLORS[TrajFlag.MACH])
            ax.text((p.distance >> PreferredUnits.distance) + max_range / 100, df['height'].min(),
                    "Mach 1", fontsize=font_size, rotation=90, color=PLOT_COLORS[TrajFlag.MACH])

    # Transparent figure and axes background
    fig.patch.set_alpha(0.0)  # Set the figure (background) to transparent
    ax.patch.set_alpha(0.0)  # Set the axis background to transparent

    max_range_in_drop_units = PreferredUnits.distance(df.distance.max()) >> PreferredUnits.drop
    # Sight line
    x_sight = [0, max_range]
    y_sight = [0, max_range_in_drop_units * math.tan(look_angle >> Angular.Radian)]
    ax.plot(x_sight, y_sight, linestyle='--', color=PLOT_COLORS['sight'])
    # Barrel pointing line
    x_bbl = [0, max_range]
    y_bbl = [-(hit_result.shot.weapon.sight_height >> PreferredUnits.drop),
             max_range_in_drop_units * math.tan(hit_result.trajectory[0].angle >> Angular.Radian)
             - (hit_result.shot.weapon.sight_height >> PreferredUnits.drop)]
    ax.plot(x_bbl, y_bbl, linestyle=':', color=PLOT_COLORS['barrel'])
    # Line labels
    sight_above_bbl = y_sight[1] > y_bbl[1]
    if (x_sight[1] - x_sight[0]) == 0:
        angle = 90
    else:
        angle = math.degrees(math.atan((y_sight[1] - y_sight[0]) / (x_sight[1] - x_sight[0])))
    ax.text(x_sight[1], y_sight[1], "Sight line", linespacing=1.2,
            rotation=angle, rotation_mode='anchor', transform_rotates_text=True,
            fontsize=font_size, color=PLOT_COLORS['sight'], ha='right',
            va='bottom' if sight_above_bbl else 'top')
    if (x_bbl[1] - x_bbl[0]) == 0:
        angle = 90
    else:
        angle = math.degrees(
            math.atan((y_bbl[1] - y_bbl[0]) / (x_bbl[1] - x_bbl[0]))
        )
    ax.text(x_bbl[1], y_bbl[1],
            "Barrel pointing", linespacing=1.2,
            rotation=angle, rotation_mode='anchor',
            transform_rotates_text=True,
            fontsize=font_size, color=PLOT_COLORS['barrel'],
            ha='right', va='top' if sight_above_bbl else 'bottom')
    # Plot velocity (on secondary axis)
    df.plot(x='distance', xlabel=PreferredUnits.distance.symbol,
            y=['velocity'], ylabel=PreferredUnits.velocity.symbol,
            secondary_y=True, color=PLOT_COLORS['velocity'],
            ylim=[0, df['velocity'].max()], ax=ax)
    # Let secondary shine through
    ax.set_zorder(1)
    ax.set_facecolor(PLOT_COLORS['face'])

    # Set frame (border) color to rgb(215, 155, 0)
    for spine in ax.spines.values():
        spine.set_edgecolor(PLOT_COLORS['frame'])
        spine.set_linewidth(1)  # Optional: set the thickness of the frame

    # Set axis labels to the same color (rgb(215, 155, 0))
    ax.xaxis.label.set_color(PLOT_COLORS['frame'])  # X-axis label color
    ax.yaxis.label.set_color(PLOT_COLORS['frame'])  # Y-axis label color
    # Secondary Y-axis label color (if applicable)
    ax.right_ax.yaxis.label.set_color(PLOT_COLORS['frame'])

    # Set the ticks color to match the frame and labels (optional)
    ax.tick_params(axis='x', colors=PLOT_COLORS['frame'])
    ax.tick_params(axis='y', colors=PLOT_COLORS['frame'])
    # For the secondary y-axis
    ax.right_ax.tick_params(axis='y', colors=PLOT_COLORS['frame'])

    if show_time_axis:
        if not backward_bending_trajectory:
            add_time_of_flight_axis(ax, hit_result)
        else:
            warnings.warn("The trajectory is backward bending. Please add custom time visualization if needed")

    return ax



"""Module for Weapon and Ammo properties definitions"""
import math
from dataclasses import dataclass

from typing_extensions import NamedTuple, Union, Optional, Any, Literal, get_args

from py_ballisticcalc.drag_model import DragModel
from py_ballisticcalc.unit import Velocity, Temperature, Distance, Angular, PreferredUnits

TrajectoryData: Any

SightFocalPlane = Literal['FFP', 'SFP', 'LWIR']


class SightReticleStep(NamedTuple):
    """Reticle step"""

    vertical: Angular
    horizontal: Angular


class SightClicks(NamedTuple):
    """SightClicks tuple"""

    vertical: float
    horizontal: float


@dataclass
class Sight:
    """Sight data for sight specific adjustment calculation"""

    focal_plane: SightFocalPlane
    scale_factor: Distance
    h_click_size: Angular
    v_click_size: Angular

    # def __post_init__(self):
    def __init__(self,
                 focal_plane: SightFocalPlane = 'FFP',
                 scale_factor: Optional[Union[float, Distance]] = None,
                 h_click_size: Optional[Union[float, Angular]] = None,
                 v_click_size: Optional[Union[float, Angular]] = None):

        if focal_plane not in get_args(SightFocalPlane):
            raise ValueError("Wrong focal plane")

        if not scale_factor and focal_plane == 'SFP':
            raise ValueError('Scale_factor required for SFP sights')

        if (
                not isinstance(h_click_size, (Angular, float, int))
                or not isinstance(v_click_size, (Angular, float, int))
        ):
            raise TypeError("type Angular expected for 'h_click_size' and 'v_click_size'")

        self.focal_plane = focal_plane
        self.scale_factor = PreferredUnits.distance(scale_factor or 1)
        self.h_click_size = PreferredUnits.adjustment(h_click_size)
        self.v_click_size = PreferredUnits.adjustment(v_click_size)

        if self.h_click_size.raw_value <= 0 or self.v_click_size.raw_value <= 0:
            raise TypeError("'h_click_size' and 'v_click_size' have to be positive")

    def _adjust_sfp_reticle_steps(self, target_distance: Union[float, Distance],
                                  magnification: float) -> SightReticleStep:
        """Calculates the SFP reticle steps for a target distance and magnification"""

        assert self.focal_plane == 'SFP', "SFP focal plane required"

        # adjust reticle scale relative to target distance and magnification
        def get_sfp_step(click_size: Angular):
            # Don't need distances conversion cause of it's destroying there
            return click_size.units(
                click_size.unit_value
                * self.scale_factor.raw_value
                / _td.raw_value
                * magnification
            )

        _td = PreferredUnits.distance(target_distance)
        _h_step = get_sfp_step(self.h_click_size)
        _v_step = get_sfp_step(self.v_click_size)
        return SightReticleStep(_h_step, _v_step)

    def get_adjustment(self, target_distance: Distance,
                       drop_adj: Angular, windage_adj: Angular,
                       magnification: float):
        """Calculate adjustment for target distance and magnification"""

        if self.focal_plane == 'SFP':
            steps = self._adjust_sfp_reticle_steps(target_distance, magnification)
            return SightClicks(
                drop_adj.raw_value / steps.vertical.raw_value,
                windage_adj.raw_value / steps.horizontal.raw_value
            )
        if self.focal_plane == 'FFP':
            return SightClicks(
                drop_adj.raw_value / self.v_click_size.raw_value,
                windage_adj.raw_value / self.h_click_size.raw_value
            )
        if self.focal_plane == 'LWIR':
            # adjust clicks to magnification
            return SightClicks(
                drop_adj.raw_value / (self.v_click_size.raw_value / magnification),
                windage_adj.raw_value / (self.h_click_size.raw_value / magnification)
            )
        raise AttributeError("Wrong focal_plane")

    def get_trajectory_adjustment(self, trajectory_point: 'TrajectoryData', magnification: float) -> SightClicks:
        """Calculate adjustment for target distance and magnification for `TrajectoryData` instance"""

        return self.get_adjustment(trajectory_point.distance,
                                   trajectory_point.drop_adj,
                                   trajectory_point.windage_adj,
                                   magnification)


@dataclass
class Weapon:
    """
    A base class for creating Weapon.

    Attributes:
        sight_height: Sight height
        twist: Twist
        zero_elevation: Zero elevation
        sight: Sight properties
    """

    sight_height: Distance
    twist: Distance
    zero_elevation: Angular
    sight: Optional[Sight]

    def __init__(self,
                 sight_height: Optional[Union[float, Distance]] = None,
                 twist: Optional[Union[float, Distance]] = None,
                 zero_elevation: Optional[Union[float, Angular]] = None,
                 sight: Optional[Sight] = None):
        """
        Create a new weapon instance with given parameters

        Args:
            sight_height: Vertical distance from center of bore line to center of sight line.
            twist: Distance for barrel rifling to complete one complete turn.
                Positive value => right-hand twist, negative value => left-hand twist.
            zero_elevation: Angle of barrel relative to sight line when sight is set to "zero."
                (Typically computed by ballistic Calculator.)
            sight: Sight properties

        Example:
            This is how you can create a weapon

            ```python
            from py_ballisticcalc import Weapon, Unis, Sight

            weapon = Weapon(
                sight_height=Unit.Inch(2.),
                twist=Unit.Inch(10.),
                zero_elevation=Unit.Mil(0),
                sight=Sight(
                    'FFP', 2,
                    h_click_size=Unit.Mil(0.2),
                    v_click_size=Unit.Mil(0.2)
                )
            )
            ```
        """
        self.sight_height = PreferredUnits.sight_height(sight_height or 0)
        self.twist = PreferredUnits.twist(twist or 0)
        self.zero_elevation = PreferredUnits.angular(zero_elevation or 0)
        self.sight = sight


@dataclass
class Ammo:
    """
    A base class for creating Weapon.

    Attributes:
        dm: DragModel for projectile
        mv: Muzzle Velocity
        powder_temp: Baseline temperature that produces the given mv
        temp_modifier: Change in velocity w temperature: % per 15°C.
            Can be computed with .calc_powder_sens().  Only applies if:
                Settings.use_powder_sensitivity = True
        use_powder_sensitivity: Flag to allow to adjust muzzle velocity to the powder sensitivity

    """

    dm: DragModel
    mv: Velocity
    powder_temp: Temperature
    temp_modifier: float
    use_powder_sensitivity: bool = False

    def __init__(self,
                 dm: DragModel,
                 mv: Union[float, Velocity],
                 powder_temp: Optional[Union[float, Temperature]] = None,
                 temp_modifier: float = 0,
                 use_powder_sensitivity: bool = False):
        """
        Create a new ammo instance with given parameters

        Args:
            dm: drag model
            mv: muzzle velocity at given powder temperature
            powder_temp: powder temperature
            temp_modifier: Change in velocity w temperature: % per 15°C.
                Can be computed with .calc_powder_sens().  Only applies if:
                Ammo.use_powder_sensitivity = True
            use_powder_sensitivity: should adjust muzzle velocity using powder sensitivity

        Example:
            This is how you can create a weapon

            ```python
            from py_ballisticcalc import Ammo, Unit, DragModel

            ammo = Ammo(
                dm=DragModel(
                    bc=0.381,
                    drag_table=TableG7,
                    weight=Unit.Grain(300),
                    length=Unit.Inch(1.7),
                    diameter=Unit.Inch(0.338),
                ),
                mv=Unit.MPS(815),
                powder_temp=Unit.Celsius(15),
                temp_modifier=0.123,
                use_powder_sensitivity=True,
            )
            ```
        """
        self.dm = dm
        self.mv = PreferredUnits.velocity(mv or 0)
        self.powder_temp = PreferredUnits.temperature(powder_temp or Temperature.Celsius(15))
        self.temp_modifier = temp_modifier or 0
        self.use_powder_sensitivity = use_powder_sensitivity

    def calc_powder_sens(self, other_velocity: Union[float, Velocity],
                         other_temperature: Union[float, Temperature]) -> float:
        """Calculates velocity correction by temperature change; assigns to self.temp_modifier

        Args:
            other_velocity: other velocity at other_temperature
            other_temperature: other temperature

        Returns:
            temperature modifier in terms %v_delta/15°C

        Example:
            ```python
            powder_sensitivity = ammo.calc_powder_sens(
                Unit.MPS(830),
                Unit.Celsius(200)
            )
            ```
        """
        v0 = self.mv >> Velocity.MPS
        t0 = self.powder_temp >> Temperature.Celsius
        v1 = PreferredUnits.velocity(other_velocity) >> Velocity.MPS
        t1 = PreferredUnits.temperature(other_temperature) >> Temperature.Celsius

        v_delta = math.fabs(v0 - v1)
        t_delta = math.fabs(t0 - t1)
        v_lower = v1 if v1 < v0 else v0

        if v_delta == 0 or t_delta == 0:
            raise ValueError(
                "Temperature modifier error, other velocity"
                " and temperature can't be same as default"
            )
        self.temp_modifier = v_delta / t_delta * (15 / v_lower)  # * 100
        return self.temp_modifier

    def get_velocity_for_temp(self, current_temp: Union[float, Temperature]) -> Velocity:
        """Calculates muzzle velocity at temperature, based on temp_modifier.

        Args:
            current_temp: Temperature of cartridge powder

        Returns:
            Muzzle velocity corrected to current_temp

        Example:
            ```python
            muzzle_velocity = ammo.get_velocity_for_temp(
                Unit.Celsius(200)
            )
            ```
        """
        if not self.use_powder_sensitivity:
            return self.mv
        try:
            v0 = self.mv >> Velocity.MPS
            t0 = self.powder_temp >> Temperature.Celsius
            t1 = PreferredUnits.temperature(current_temp) >> Temperature.Celsius
            t_delta = t1 - t0
            muzzle_velocity = self.temp_modifier / (15 / v0) * t_delta + v0
        except ZeroDivisionError:
            muzzle_velocity = 0
        return Velocity.MPS(muzzle_velocity)


__all__ = ('Weapon', 'Ammo', 'Sight', 'SightFocalPlane', 'SightClicks', 'SightReticleStep')



"""Implements a point of trajectory class in applicable data types"""
from dataclasses import dataclass, field

from typing_extensions import NamedTuple, Optional, Union, Any, Tuple, Final

from py_ballisticcalc.conditions import Shot
from py_ballisticcalc.unit import Angular, Distance, Weight, Velocity, Energy, GenericDimension, Unit, PreferredUnits

__all__ = ('TrajectoryData', 'HitResult', 'TrajFlag', 'DangerSpace')

DataFrame: Any  # Disable type-checking for DataFrames
Axes: Any       # Disable type-checking for Axes

_TrajFlagNames = {
    0: 'NONE',
    1: 'ZERO_UP',
    2: 'ZERO_DOWN',
    3: 'ZERO',
    4: 'MACH',
    8: 'RANGE',
    16: 'APEX',
    31: 'ALL',
}


class TrajFlag(int):
    """Flags for marking trajectory row if Zero or Mach crossing.
    Also used to set filters for a trajectory calculation loop.
    """
    NONE: Final[int] = 0
    ZERO_UP: Final[int] = 1
    ZERO_DOWN: Final[int] = 2
    ZERO: Final[int] = ZERO_UP | ZERO_DOWN
    MACH: Final[int] = 4
    RANGE: Final[int] = 8
    APEX: Final[int] = 16
    ALL: Final[int] = RANGE | ZERO_UP | ZERO_DOWN | MACH | APEX

    @staticmethod
    def name(value: Union[int, 'TrajFlag']) -> str:
        """Return a concatenated name representation of the given flag value."""
        if value in _TrajFlagNames:
            return _TrajFlagNames[value]

        parts = [name for bit, name in _TrajFlagNames.items() if bit and (value & bit) == bit]
        if "ZERO_UP" in parts and "ZERO_DOWN" in parts:
            parts.remove("ZERO_UP")
            parts.remove("ZERO_DOWN")
        return "|".join(parts) if parts else "UNKNOWN"


class TrajectoryData(NamedTuple):
    """
    Data for one point in ballistic trajectory.

    Attributes:
        time (float): Bullet flight time in seconds.
        distance (Distance): Down-range (x-axis) coordinate of this point.
        velocity (Velocity): Velocity.
        mach (float): Velocity in Mach terms.
        height (Distance): Vertical (y-axis) coordinate of this point.
        target_drop (Distance): Drop relative to sight-line.
        drop_adj (Angular): Sight adjustment to zero target_drop at this distance.
        windage (Distance): Windage (z-axis) coordinate of this point.
        windage_adj (Angular): Windage adjustment.
        look_distance (Distance): Sight-line distance = .distance/cosine(look_angle).
        angle (Angular): Angle of velocity vector relative to x-axis.
        density_factor (float): Ratio of air density here to standard density.
        drag (float): Current drag coefficient.
        energy (Energy): Energy of bullet at this point.
        ogw (Weight): Optimal game weight, given .energy.
        flag (Union[TrajFlag, int]): Row type.
    """

    time: float
    distance: Distance
    velocity: Velocity
    mach: float
    height: Distance
    target_drop: Distance
    drop_adj: Angular
    windage: Distance
    windage_adj: Angular
    look_distance: Distance
    angle: Angular
    density_factor: float
    drag: float
    energy: Energy
    ogw: Weight
    flag: Union[TrajFlag, int]

    def formatted(self) -> Tuple[str, ...]:
        """
        Returns:
            Tuple[str, ...]: Matrix of formatted strings for this point, in PreferredUnits.
        """

        def _fmt(v: GenericDimension, u: Unit) -> str:
            """simple formatter"""
            return f"{v >> u:.{u.accuracy}f} {u.symbol}"

        return (
            f'{self.time:.3f} s',
            _fmt(self.distance, PreferredUnits.distance),
            _fmt(self.velocity, PreferredUnits.velocity),
            f'{self.mach:.2f} mach',
            _fmt(self.height, PreferredUnits.drop),
            _fmt(self.target_drop, PreferredUnits.drop),
            _fmt(self.drop_adj, PreferredUnits.adjustment),
            _fmt(self.windage, PreferredUnits.drop),
            _fmt(self.windage_adj, PreferredUnits.adjustment),
            _fmt(self.look_distance, PreferredUnits.distance),
            _fmt(self.angle, PreferredUnits.angular),
            f'{self.density_factor:.3e}',
            f'{self.drag:.3f}',
            _fmt(self.energy, PreferredUnits.energy),
            _fmt(self.ogw, PreferredUnits.ogw),
            TrajFlag.name(self.flag)
        )

    def in_def_units(self) -> Tuple[float, ...]:
        """
        Returns:
            Tuple[float, ...]: Matrix of floats of this point, in PreferredUnits.
        """
        return (
            self.time,
            self.distance >> PreferredUnits.distance,
            self.velocity >> PreferredUnits.velocity,
            self.mach,
            self.height >> PreferredUnits.drop,
            self.target_drop >> PreferredUnits.drop,
            self.drop_adj >> PreferredUnits.adjustment,
            self.windage >> PreferredUnits.drop,
            self.windage_adj >> PreferredUnits.adjustment,
            self.look_distance >> PreferredUnits.distance,
            self.angle >> PreferredUnits.angular,
            self.density_factor,
            self.drag,
            self.energy >> PreferredUnits.energy,
            self.ogw >> PreferredUnits.ogw,
            self.flag
        )


class DangerSpace(NamedTuple):
    """Stores the danger space data for distance specified.

    Attributes:
        at_range (TrajectoryData): Trajectory data at the specified range.
        target_height (Distance): Target height.
        begin (TrajectoryData): Beginning of danger space.
        end (TrajectoryData): End of danger space.
        look_angle (Angular): Look-angle (sight-line).
    """
    at_range: TrajectoryData
    target_height: Distance
    begin: TrajectoryData
    end: TrajectoryData
    look_angle: Angular

    def __str__(self) -> str:
        return f'Danger space at {self.at_range.distance << PreferredUnits.distance} ' \
            + f'for {self.target_height << PreferredUnits.drop} tall target ' \
            + (f'at {self.look_angle << Angular.Degree} look-angle ' if self.look_angle != 0 else '') \
            + f'ranges from {self.begin.distance << PreferredUnits.distance} ' \
            + f'to {self.end.distance << PreferredUnits.distance}'

    # pylint: disable=import-outside-toplevel
    def overlay(self, ax: 'Axes', label: Optional[str] = None):  # type: ignore
        """Highlights danger-space region on plot.

        Args:
            ax (Axes): The axes to overlay on.
            label (Optional[str], optional): Label for the overlay. Defaults to None.

        Raises:
            ImportError: If plotting dependencies are not installed.
        """
        try:
            from py_ballisticcalc.visualize.plot import add_danger_space_overlay
            add_danger_space_overlay(self, ax, label)
        except ImportError as err:
            raise ImportError(
                "Use `pip install py_ballisticcalc[charts]` to get results as a plot"
            ) from err


# pylint: disable=import-outside-toplevel
@dataclass(frozen=True)
class HitResult:
    """Computed trajectory data of the shot.

    Attributes:
        shot (Shot): The shot conditions.
        trajectory (list[TrajectoryData]): The trajectory data.
        extra (bool): Whether special points (TrajFlag.ALL) were requested.
    """
    shot: Shot
    trajectory: list[TrajectoryData] = field(repr=False)
    extra: bool = False

    def __iter__(self):
        yield from self.trajectory

    def __getitem__(self, item):
        return self.trajectory[item]

    def __check_extra__(self):
        if not self.extra:
            raise AttributeError(
                f"{object.__repr__(self)} has no extra data. "
                f"Use Calculator.fire(..., extra_data=True)"
            )

    def zeros(self) -> list[TrajectoryData]:
        """
        Returns:
            list[TrajectoryData]: Zero crossing points.

        Raises:
            AttributeError: If extra_data was not requested.
            ArithmeticError: If zero crossing points are not found.
        """
        self.__check_extra__()
        data = [row for row in self.trajectory if row.flag & TrajFlag.ZERO]
        if len(data) < 1:
            raise ArithmeticError("Can't find zero crossing points")
        return data

    def index_at_distance(self, d: Distance) -> int:
        """
        Args:
            d (Distance): Distance for which we want Trajectory Data.

        Returns:
            int: Index of first trajectory row with .distance >= d; otherwise -1.
        """
        # Get index of first trajectory point with distance >= at_range
        epsilon = 1e-8  # small value to avoid floating point issues
        return next((i for i in range(len(self.trajectory))
                     if self.trajectory[i].distance.raw_value >= d.raw_value - epsilon), -1)

    def get_at_distance(self, d: Distance) -> TrajectoryData:
        """
        Args:
            d (Distance): Distance for which we want Trajectory Data.

        Returns:
            TrajectoryData: First trajectory row with .distance >= d.

        Raises:
            ArithmeticError: If trajectory doesn't reach requested distance.
        """
        if (i := self.index_at_distance(d)) < 0:
            raise ArithmeticError(
                f"Calculated trajectory doesn't reach requested distance {d}"
            )
        return self.trajectory[i]

    def danger_space(self,
                     at_range: Union[float, Distance],
                     target_height: Union[float, Distance],
                     look_angle: Optional[Union[float, Angular]] = None
                     ) -> DangerSpace:
        """
        Calculates the danger space for a given range and target height.

        Assumes that the trajectory hits the center of a target at any distance.
        Determines how much ranging error can be tolerated if the critical region
        of the target has height *h*. Finds how far forward and backward along the
        line of sight a target can move such that the trajectory is still within *h*/2
        of the original drop.

        Args:
            at_range (Union[float, Distance]): Danger space is calculated for a target centered at this sight distance.
            target_height (Union[float, Distance]): Target height (*h*) determines danger space.
            look_angle (Optional[Union[float, Angular]], optional): Ranging errors occur along the look angle to the target.

        Returns:
            DangerSpace: The calculated danger space.

        Raises:
            AttributeError: If extra_data wasn't requested.
            ArithmeticError: If trajectory doesn't reach requested distance.
        """
        self.__check_extra__()

        at_range = PreferredUnits.distance(at_range)
        target_height = PreferredUnits.distance(target_height)
        target_height_half = target_height.raw_value / 2.0

        _look_angle: Angular
        if look_angle is None:
            _look_angle = self.shot.look_angle
        else:
            _look_angle = PreferredUnits.angular(look_angle)

        # Get index of first trajectory point with distance >= at_range
        if (index := self.index_at_distance(at_range)) < 0:
            raise ArithmeticError(
                f"Calculated trajectory doesn't reach requested distance {at_range}"
            )

        def find_begin_danger(row_num: int) -> TrajectoryData:
            """
            Beginning of danger space is last .distance' < .distance where
                (.drop' - target_center) >= target_height/2

            Args:
                row_num (int): Index of the trajectory point for which we are calculating danger space.

            Returns:
                TrajectoryData: Distance marking beginning of danger space.
            """
            center_row = self.trajectory[row_num]
            for prime_row in reversed(self.trajectory[:row_num]):
                if (prime_row.target_drop.raw_value - center_row.target_drop.raw_value) >= target_height_half:
                    return prime_row
            return self.trajectory[0]

        def find_end_danger(row_num: int) -> TrajectoryData:
            """
            End of danger space is first .distance' > .distance where
                (target_center - .drop') >= target_height/2

            Args:
                row_num (int): Index of the trajectory point for which we are calculating danger space.

            Returns:
                TrajectoryData: Distance marking end of danger space.
            """
            center_row = self.trajectory[row_num]
            for prime_row in self.trajectory[row_num + 1:]:
                if (center_row.target_drop.raw_value - prime_row.target_drop.raw_value) >= target_height_half:
                    return prime_row
            return self.trajectory[-1]

        return DangerSpace(self.trajectory[index],
                           target_height,
                           find_begin_danger(index),
                           find_end_danger(index),
                           _look_angle)

    def dataframe(self, formatted: bool = False) -> 'DataFrame':  # type: ignore
        """
        Returns the trajectory table as a DataFrame.

        Args:
            formatted (bool, optional): False for values as floats; True for strings with prefer_units. Defaults to False.

        Returns:
            DataFrame: The trajectory table as a DataFrame.

        Raises:
            ImportError: If pandas or plotting dependencies are not installed.
        """
        try:
            from py_ballisticcalc.visualize.dataframe import hit_result_as_dataframe
            return hit_result_as_dataframe(self, formatted)
        except ImportError as err:
            raise ImportError(
                "Use `pip install py_ballisticcalc[charts]` to get trajectory as pandas.DataFrame"
            ) from err

    def plot(self, look_angle: Optional[Angular] = None) -> 'Axes':  # type: ignore
        """
        Returns a graph of the trajectory.

        Args:
            look_angle (Optional[Angular], optional): Look angle for the plot. Defaults to None.

        Returns:
            Axes: The plot axes.

        Raises:
            ImportError: If plotting dependencies are not installed.
        """
        try:
            from py_ballisticcalc.visualize.plot import hit_result_as_plot
            return hit_result_as_plot(self, look_angle)
        except ImportError as err:
            raise ImportError(
                "Use `pip install py_ballisticcalc[charts]` to get results as a plot"
            ) from err



"""Example of library usage"""

# pylint: disable=wildcard-import,unused-wildcard-import
from py_ballisticcalc import *

# Modify default prefer_units
PreferredUnits.velocity = Velocity.FPS
PreferredUnits.temperature = Temperature.Celsius
PreferredUnits.distance = Distance.Meter
PreferredUnits.sight_height = Distance.Centimeter

# Define ammunition parameters
weight, diameter = 168, 0.308  # Numbers will be assumed to use default Settings.Units
length: Distance = Distance.Inch(1.282)  # Or declare prefer_units explicitly
dm = DragModel(0.223, TableG7, weight, diameter, length)
ammo = Ammo(dm, 2750, 15, use_powder_sensitivity=True)
ammo.calc_powder_sens(2723, 0)
gun = Weapon(sight_height=9, twist=12)
current_atmo = Atmo(110, 29.8, 15, 72)
current_winds = [Wind(2, 90)]
shot = Shot(weapon=gun, ammo=ammo, atmo=current_atmo, winds=current_winds)
calc: Calculator = Calculator()
calc.set_weapon_zero(shot, Distance.Meter(100))

shot_result = calc.fire(shot, trajectory_range=1000, trajectory_step=100)

for p in shot_result:
    print(p.formatted())



import math
from typing import Union, NamedTuple

__all__ = ('Vector',)


class Vector(NamedTuple):
    """
    Attributes:
        x (float): Distance component.
        y (float): Vertical component.
        z (float): Horizontal component.
    """

    x: float
    y: float
    z: float

    def magnitude(self) -> float:
        """
        Returns:
            magnitude of Vector instance
        """
        # return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)
        return math.hypot(self.x, self.y, self.z)

    def mul_by_const(self, a: float) -> 'Vector':
        """
        Args:
            a: float constant
        Returns:
            Vector instance
        """
        return Vector(self.x * a, self.y * a, self.z * a)

    def mul_by_vector(self, b: 'Vector') -> float:
        """
        Args:
            b: other Vector instance
        Returns:
            float result of multiplication of two Vector instances
        """
        return self.x * b.x + self.y * b.y + self.z * b.z

    def add(self, b: 'Vector') -> 'Vector':
        """
        Args:
            b: other Vector instance
        Returns:
            Vector instance - result of adding of two Vector instances
        """
        return Vector(self.x + b.x, self.y + b.y, self.z + b.z)

    def subtract(self, b: 'Vector') -> 'Vector':
        """
        Args:
            b: other Vector instance
        Returns:
            Vector instance - result of subtract of two Vector instances
        """
        return Vector(self.x - b.x, self.y - b.y, self.z - b.z)

    def negate(self) -> 'Vector':
        """
        Returns:
            Vector instance negative to current
        """
        return Vector(-self.x, -self.y, -self.z)

    def normalize(self) -> 'Vector':
        """
        Returns:
            Normalized Vector instance
        """
        m = self.magnitude()
        if math.fabs(m) < 1e-10:
            return Vector(self.x, self.y, self.z)
        return self.mul_by_const(1.0 / m)

    def __mul__(self, other: Union[int, float, 'Vector']) -> Union[float, 'Vector']:  # type: ignore[override]
        if isinstance(other, (int, float)):
            return self.mul_by_const(other)
        if isinstance(other, Vector):
            return self.mul_by_vector(other)
        raise TypeError(other)

    # aliases more efficient than wrappers
    def __add__(self, other: 'Vector') -> 'Vector':  # type: ignore[override]
        return self.add(other)

    def __radd__(self, other: 'Vector') -> 'Vector':  # type: ignore[override]
        return self.add(other)

    def __iadd__(self, other: 'Vector') -> 'Vector':  # type: ignore[override]
        return self.add(other)

    def __sub__(self, other: 'Vector') -> 'Vector':  # type: ignore[override]
        return self.subtract(other)

    def __rsub__(self, other: 'Vector') -> 'Vector':  # type: ignore[override]
        return self.subtract(other)

    def __isub__(self, other: 'Vector') -> 'Vector':  # type: ignore[override]
        return self.subtract(other)

    def __rmul__(self, other: Union[int, float, 'Vector']) -> Union[float, 'Vector']:  # type: ignore[override]
        return self.__mul__(other)

    def __imul__(self, other: Union[int, float, 'Vector']) -> Union[float, 'Vector']:  # type: ignore[override]
        return self.__mul__(other)

    def __neg__(self) -> 'Vector':  # type: ignore[override]
        return self.negate()
